import datetime
import os

from google.cloud import datastore

from .dicplus import DicPlus

if not os.getenv('GOOGLE_APPLICATION_CREDENTIALS') or not os.getenv('GOOGLE_CLOUD_PROJECT'):
    raise Exception("No GOOGLE_APPLICATION_CREDENTIALS")

ds = datastore.Client(project=os.getenv('GOOGLE_CLOUD_PROJECT'))


class Store():
    def __init__(self, kind, lookup=None, default_entity=None):
        self.kind = kind
        if default_entity != None:
            self.entity = default_entity
        elif lookup is None:
            newkey = ds.key(kind)
            self.entity = datastore.Entity(key=newkey)
            self.update({"created": datetime.datetime.utcnow()})
        else:
            lookup_k = lookup.split(':')[0]
            lookup_v = lookup.split(':')[1]
            if lookup_k == "id":
                mykey = ds.key(kind, int(lookup_v))
                self.entity = ds.get(mykey)
            else:
                query = ds.query(kind=kind)
                query.add_filter(lookup_k, "=", lookup_v)
                found = list(query.fetch())
                if len(found) > 0:
                    self.entity = list(query.fetch())[0]
                else:
                    self.entity = None
            if self.entity is None:
                self.entity = datastore.Entity(key=ds.key(kind))
                self.update(
                    {lookup_k: lookup_v, "created": datetime.datetime.utcnow()})

    def update(self, dic):
        self.entity.update(dic)
        self.save()

    @property
    def key_id(self):
        return self.entity.key.id

    def empty(self):
        return (not 'created' in dict(self.entity))

    def __str__(self):
        return str(dict(self.entity))

    def save(self):
        ds.put(self.entity)

    def delete(self):
        ds.delete(self.entity.key)

    @property
    def data(self):
        return self.toDM(json_datetime=True)

    def toDP(self, json_datetime=False):
        return self.ds_to_dp(self.entity, json_datetime=json_datetime)

    def toDM(self, json_datetime=False):
        return self.ds_to_dp(self.entity, json_datetime=json_datetime)

    def ds_to_dp(self, entity, is_list=False, json_datetime=False):  # to re-format
        dp = DicPlus()
        if is_list:
            new_list = []
            for el in entity:
                new_list.append(self.ds_to_dp(el, json_datetime=json_datetime))
            return new_list
        for k, v in dict(entity).items():
            if type(v) == datastore.entity.Entity:
                dp[k] = self.ds_to_dp(v, json_datetime=json_datetime)
            elif type(v) == list:
                dp[k] = self.ds_to_dp(
                    v, is_list=True, json_datetime=json_datetime)
            elif type(v) == datetime.datetime:
                if json_datetime:
                    dp[k] = v.strftime("%Y-%m-%dT%H:%M:%S")
                else:
                    dp[k] = v.replace(tzinfo=None)
            else:
                dp[k] = v
        return dp

    @classmethod
    def findAll(cls, kind, filters):
        ''' Use this template:
        Store.findAll('message', [['test', 'test2']])
        '''
        try:
            query = ds.query(kind=kind)
            for f in filters:
                if len(f) == 2:
                    query.add_filter(f[0], "=", f[1])
                else:
                    query.add_filter(f[0], f[1], f[2])
            results = list(query.fetch())
            results_objs = []
            for obj in results:
                results_objs.append(cls(kind, default_entity=obj))
            return results_objs
        except Exception as e:
            return None

    @classmethod
    def findOne(cls, kind, filters):
        ''' Use this template:
        Store.findAll('message', [['test', 'test2']])
        '''
        list_found = cls.findAll(kind, filters)
        if list_found == None:
            return None
        elif len(list_found) > 0:
            return list_found[0]
        else:
            return None



import hashlib
import json
import os
import socket
import time
import uuid
from base64 import b64encode
from calendar import timegm
from datetime import datetime, timedelta
from functools import wraps

import requests
from jose import jwk, jws, jwt

from flask import g, request

from .dicplus import DicPlus

ALGORITHMS = ["RS256"]
EXPIRATION = 1000


def hash_digest(body):
    myhash = hashlib.sha512()
    myhash.update(body)
    digest = b64encode(myhash.digest()).decode()
    return digest


def get_token_header(auth_type, request):
    """Obtains the Access Token from the Authorization Header
    """
    auth = request.headers.get("Authorization", None)
    if not auth:
        raise Exception("No Authorization Header")

    auth_parts = auth.split()

    if auth_parts[0].lower() != auth_type.lower():
        raise Exception(
            "Authorization Header must start with {}".format(auth_type))
    elif len(auth_parts) == 1:
        raise Exception("No Signature attached")
    elif len(auth_parts) > 2:
        raise Exception("Invalid Header")
    token = auth_parts[1]

    return token


class Secure():
    def __init__(self, service, private_key, keySet=None, signer=None, authenticator=None):
        self.service = service
        self.keySet = keySet
        self.private_key = private_key
        if authenticator:
            self.authenticator = Authenticator(
                client_id=authenticator['client_id'],
                client_secret=authenticator['client_secret'],
                audience=self.service,
                domain=authenticator['domain']
            )
        self.authorizer = Authorizer(audience=self.service, keySet=self.keySet)
        self.signer = Signer(issuer=signer or self.service,
                             private_key=self.private_key)
        self.verifyer = Signature_Verifyer(
            audience=self.service, keySet=self.keySet)

    def validate_scope(self, scope, token=None):
        if not token:
            token_payload = g.authorize_token_payload
        else:
            token_payload = jwt.get_unverified_claims(token)
        if scope == None or scope == "":
            return True
        if token_payload.get("scope"):
            scope = scope.split()
            for s in scope:
                if s not in token_payload['scope']:
                    return False
        return True

    def authorize(self, scope, issuers=None):
        def inner_authorize(f):
            @wraps(f)
            def decorated(*args, **kwargs):
                try:
                    token = get_token_header("Bearer", request)
                    payload = self.authorizer.decode(token, issuers)
                    g.authorize_token_payload = payload
                except:
                    raise Exception("Invalid Header", 401)
                if not self.validate_scope(scope, token):
                    raise Exception("Invalid Scope", 401)
                return f(*args, **kwargs)
            return decorated
        return inner_authorize

    def get_access_token(self):
        return self.authenticator.access_token

    def verify(self, issuers=None):
        def inner_verify(f):
            @wraps(f)
            def decorated(*args, **kwargs):
                # Assumes the request is available
                token = get_token_header("signature", request)
                body = request.get_data()
                if not self.verifyer.verify(token, body, issuers):
                    raise Exception("Wrong Signature", 401)

                return f(*args, **kwargs)

            return decorated
        return inner_verify

    def sign(self, body, audience):
        return self.signer.sign(body, audience)

    def signed_headers(self, body, audience):
        headers = {
            "Authorization": "Signature {}".format(self.sign(body, audience))
        }
        return headers

    def auth_headers(self):
        headers = {
            "Authorization": "Bearer {}".format(self.authenticator.access_token)
        }
        return headers


class KeySet():
    def __init__(self, jsonfile=None):
        self.issuers = []
        if jsonfile:
            self.from_json(jsonfile)

    def add_issuer(self, issuer, keys=None):
        if not keys:
            try:
                url = issuer+".well-known/jwks.json"
                keys = requests.get(url=url).json()['keys']
            except:
                raise Exception(
                    "No keys provided nor found on {}.well-known/jwks.json".format(issuer))
        if type(keys) != list:
            keys = [keys]
        self.issuers.append({"issuer": issuer, "keys": keys})

    def from_json(self, jsonfile):
        dp = DicPlus.from_json(jsonfile)
        for ks in dp.keySet:
            issuer = ks.toDict()
            if "keys" not in ks:
                self.add_issuer(issuer["issuer"])
            else:
                self.add_issuer(issuer["issuer"], issuer["keys"])

    def get_keys(self, issuers=None):
        keys = []
        if type(issuers) != list and issuers is not None:
            issuers = [issuers]
        for i in self.issuers:
            if issuers:
                if i['issuer'] not in issuers:
                    continue
            for key in i['keys']:
                keys.append(key)

        return {"keys": keys}


class Authenticator():
    # Using Auth0
    def __init__(self, domain, client_id, client_secret, audience):
        self.domain = domain
        self.client_id = client_id
        self.client_secret = client_secret
        self.audience = audience
        self.getAccessToken()
        if self._access_token is None:
            raise Exception("Request for access token failed")

    def getAccessToken(self):
        body = {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "audience": self.audience,
            "grant_type": "client_credentials"
        }
        try:
            response = requests.post(
                url="{}/oauth/token".format(self.domain), json=body)
        except Exception as e:
            raise Exception(e)
            return None
        else:
            json_response = DicPlus.json_loads(response.text)
            self._access_token = json_response.access_token
            self._access_token_expiration = time.time() + float(json_response.expires_in)

    @property
    def access_token(self):
        if time.time() > self._access_token_expiration:
            self.getAccessToken()
        return self._access_token


class Authorizer():
    def __init__(self, audience, keySet):
        self.audience = audience
        self.keySet = keySet

    def decode(self, token, issuers=None):
        try:
            jwks = self.keySet.get_keys(issuers)
            payload = jwt.decode(
                token,
                jwks,
                audience=self.audience
            )
        except jwt.ExpiredSignatureError:
            raise Exception("Expired Signature", 401)
        except jwt.JWTClaimsError:
            raise Exception("Please check audience and issuer", 401)
        except Exception:
            raise Exception("Invalid header", 401)

        return payload


class Signature_Verifyer():

    def __init__(self, audience, keySet, jti_cache=None):

        self.jti_cache = jti_cache
        self.algorithms = ALGORITHMS
        self.audience = audience
        self.keySet = keySet

    def verify(self, token, body, issuers=None):
        jwks = self.keySet.get_keys(issuers)
        verified_token = DicPlus(json.loads(
            jws.verify(token, jwks, self.algorithms)))
        hashed_body = hash_digest(body)
        if verified_token.digest != hashed_body:
            raise Exception("Digest Not Equal")
        if verified_token.aud != self.audience:
            raise Exception("Audience Not Equal")
        if verified_token.exp < timegm(datetime.utcnow().utctimetuple()):
            raise Exception("Token Expired")
        return True


class Signer():

    def __init__(self, issuer, private_key, algorithm=ALGORITHMS[0], expiration=EXPIRATION):
        self.algorithm = algorithm
        self.expiration = expiration
        self.issuer = issuer
        self.private_key = private_key

    def sign(self, body, audience, is_json=False):
        jti = str(uuid.uuid4())
        enc_body = json.dumps(body, separators=(',', ': ')).encode()
        digest = hash_digest(enc_body)
        token_payload = {
            "digest": digest,
            "jti": jti,
            "iat": timegm(datetime.utcnow().utctimetuple()),
            "exp": timegm(datetime.utcnow().utctimetuple()) + self.expiration,
            "iss": self.issuer,
            "aud": audience
        }
        signed_token = jws.sign(
            token_payload, self.private_key, algorithm=self.algorithm)

        return signed_token



import hashlib
import json
import uuid
from base64 import b64encode
from calendar import timegm
from datetime import datetime, timedelta
from functools import wraps

import requests
from jose import jwk, jws, jwt

from flask import request

from .dicplus import DicPlus

ALGORITHMS = ["RS256"]


def hash_digest(body):
    myhash = hashlib.sha512()
    myhash.update(body)
    digest = b64encode(myhash.digest()).decode()
    return digest


class Verifyer ():

    def __init__(self, issuer, audience, jti_cache=None, keySet=None):

        self.issuer = str(issuer)
        self.jti_cache = jti_cache
        self.algorithms = ALGORITHMS
        self.audience = audience
        if not keySet:
            url = self.issuer+"/.well-known/jwks.json"
            self.jwks = requests.get(url=url).json()['keys']
        else:
            self.jwks = keySet

    def get_signature_token_header(self, request):
        """Obtains the Access Token from the Authorization Header
        """
        auth = request.headers.get("Authorization", None)
        if not auth:
            raise Exception("No Authorization Header")

        auth_parts = auth.split()

        if auth_parts[0].lower() != "signature":
            raise Exception("Authorization Header must start with Signature")
        elif len(auth_parts) == 1:
            raise Exception("No Signature attached")
        elif len(auth_parts) > 2:
            raise Exception("Invalid Header")
        token = auth_parts[1]

        return token

    def verify(self, f):
        @wraps(f)
        def decorated(*args, **kwargs):
            # Assumes the request is available
            try:
                token = self.get_signature_token_header(request)
                verified_token = DicPlus(json.loads(
                    jws.verify(token, self.jwks, self.algorithms)))
                hashed_body = hash_digest(request.get_data())
            except:
                raise Exception("Error occured during Verification")
            if verified_token.digest != hashed_body:
                raise Exception("Digest Not Equal")
            if verified_token.aud != self.audience:
                raise Exception("Audience Not Equal")
            if verified_token.iss != self.issuer:
                raise Exception("Issuer Not Equal")
            if verified_token.exp < timegm(datetime.utcnow().utctimetuple()):
                raise Exception("Token Expired")

            return f(*args, **kwargs)

        return decorated


class Signer():

    def __init__(self, issuer, audience, private_key, algorithm=ALGORITHMS[0], expiration=100):
        self.audience = audience
        self.algorithm = algorithm
        self.expiration = expiration
        self.issuer = issuer
        self.private_key = private_key

    def sign(self, body, is_json=False):
        jti = str(uuid.uuid4())
        enc_body = json.dumps(body, separators=(',', ': ')).encode()
        digest = hash_digest(enc_body)
        token_payload = {
            "digest": digest,
            "jti": jti,
            "iat": timegm(datetime.utcnow().utctimetuple()),
            "exp": timegm(datetime.utcnow().utctimetuple()) + self.expiration,
            "iss": self.issuer,
            "aud": self.audience
        }
        signed_token = jws.sign(
            token_payload, self.private_key, algorithm=self.algorithm)

        return signed_token
